/**
  *  \file planetscentral/console/router.cc
  *  \brief Console: Router context
  */

#include <stdexcept>
#include <iostream>
#include <sys/time.h>
#include <time.h>
#include "planetscentral/console/router.h"
#include "planetscentral/dbclient/connection.h"
#include "planetscentral/dbclient/routerclient.h"
#include "util/misc.h"
#include "cpluslib/format.h"

RouterContext::RouterContext()
{ }

void
RouterContext::enter()
{ }

void
RouterContext::leave()
{
    session.clear();
}

IntValue*
RouterContext::processCommand(StringList_t& cmd)
{
    if (cmd[0] == "repeat") {
        // Process command repeatedly, for benchmarking
        if (cmd.size() < 3) {
            throw std::runtime_error("Too few arguments");
        }
        int32_t n;
        if (!stringValue(cmd[1], n) || n <= 0) {
            throw std::runtime_error("Expecting number");
        }
        struct timeval beg, end;
        gettimeofday(&beg, 0);
        for (int32_t i = 0; i < n; ++i) {
            if (isSessionCommand(cmd, 2)) {
                processSessionCommand(session, cmd, 2);
            } else {
                processCommand(cmd, 2);
            }
        }
        gettimeofday(&end, 0);

        int32_t seconds = end.tv_sec - beg.tv_sec;
        int32_t usecs   = end.tv_usec - beg.tv_usec;
        if (usecs < 0) {
            seconds--;
            usecs += 1000000;
        }
        int32_t msecs = usecs / 1000;

        return makeStringValue(format("%d.%03d seconds (=%d ms per iteration)",
                                      seconds, msecs, (1000*seconds+msecs)/n));
    } else if (cmd[0] == "s") {
        // Select session
        if (cmd.size() <= 1) {
            throw std::runtime_error("Too few arguments");
        } else if (cmd.size() == 2) {
            std::cout << "Selected session '" << cmd[1] << "'.\n";
            session = cmd[1];
            return 0;
        } else {
            return processSessionCommand(cmd[1], cmd, 2);
        }
    } else if (isSessionCommand(cmd, 0)) {
        // command directed at session
        return processSessionCommand(session, cmd, 0);
    } else {
        // raw command
        return processCommand(cmd, 0);
    }
}

string_t
RouterContext::getPrefix()
{
    if (session.empty()) {
        return "router";
    } else {
        return "router:" + session;
    }
}

bool
RouterContext::checkConfig(string_t key, string_t value)
{
    if (key == "ROUTER.HOST") {
        client.setHost(value);
        return true;
    } else if (key == "ROUTER.PORT") {
        client.setPort(value);
        return true;
    } else {
        return false;
    }
}

bool
RouterContext::isSessionCommand(StringList_t& cmd, StringList_t::size_type index)
{
    return !session.empty()
        && cmd.size() > index
        && (cmd[index] == "help"
            || cmd[index] == "stat"
            || cmd[index] == "get"
            || cmd[index] == "post"
            || cmd[index] == "quit"
            || (cmd[index] == "save" && cmd.size() == index+1));
}

IntValue*
RouterContext::processSessionCommand(string_t s, StringList_t& cmd, StringList_t::size_type start)
{
    string_t str = "S " + s + "\n";
    for (StringList_t::size_type i = start; i < cmd.size(); ++i) {
        if (i != start) {
            str += " ";
        }
        str += cmd[i];
    }
    str += "\n";
    return callRouter(str);
}

IntValue*
RouterContext::processCommand(StringList_t& cmd, StringList_t::size_type start)
{
    string_t str;
    for (StringList_t::size_type i = start; i < cmd.size(); ++i) {
        if (i != start) {
            str += " ";
        }
        str += cmd[i];
    }
    str += "\n";
    return callRouter(str);
}

IntValue*
RouterContext::callRouter(string_t cmd)
{
    return makeStringValue(client.call(cmd));
}
