/**
  *  \file planetscentral/dbclient/reconnect.cc
  *  \brief Database Client: Automatic Reconnection
  */

#include <stdio.h>
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/dbclient/connection.h"

/** Constructor.
    \param conn Connection to manage
    \param outputMutex Mutex locking console output
    \param host Default host
    \param port Default port */
ReconnectingConnection::ReconnectingConnection(Connection& conn, Mutex& outputMutex, string_t host, string_t port)
    : mtx(),
      outputMutex(outputMutex),
      conn(conn),
      host(host),
      port(port)
{ }

/** Prepare for using the connection. Obtains the lock and reconnects if necessary. */
void
ReconnectingConnection::enter()
{
    mtx.enter();
    try {
        if (conn.maybeReconnect(host, port)) {
            MutexGuard g(outputMutex);
            printf("(reconnecting to %s:%s...)\n", host.c_str(), port.c_str());
        }
    }
    catch (...) {
        mtx.leave();
        throw;
    }
}

/** Finish using the connection. Releases the lock. */
void
ReconnectingConnection::leave()
{
    mtx.leave();
}

/** Establish initial connection. This connects to the configured
    server, and retries if necessary. */
void
ReconnectingConnection::connectRetry()
{
    conn.connectRetry(host, port);
}

/** Change host. To be called before initial connection. */
void
ReconnectingConnection::setHost(string_t h)
{
    host = h;
}

/** Change port. To be called before initial connection. */
void
ReconnectingConnection::setPort(string_t p)
{
    port = p;
}

string_t
ReconnectingConnection::getHost() const
{
    return host;
}

string_t
ReconnectingConnection::getPort() const
{
    return port;
}

/*************************** ReconnectingGuard ***************************/

/** Constructor. Obtains the lock. */
ReconnectingGuard::ReconnectingGuard(ReconnectingConnection& rc)
    : rc(rc)
{
    rc.enter();
}

/** Destructor. Releases the lock. */
ReconnectingGuard::~ReconnectingGuard()
{
    rc.leave();
}
