/**
  *  \file planetscentral/dbexport/exdb.cc
  *  \brief Database Exporter: Raw Database Export
  */

#include <memory>
#include <ostream>
#include "planetscentral/dbexport/exdb.h"
#include "planetscentral/dbclient/dbclient.h"
#include "int/args.h"
#include "planetscentral/dbexport/dbexport.h"
#include "planetscentral/dbclient/reconnect.h"
#include "cpluslib/format.h"
#include "planetscentral/dbclient/dbkey.h"
#include "planetscentral/dbclient/dbstring.h"
#include "planetscentral/dbclient/dbset.h"
#include "planetscentral/dbclient/dblist.h"
#include "planetscentral/dbclient/dbhash.h"


static string_t
quoteConsoleString(string_t arg)
{
    // Check for unprintable characters
    bool isPrintable = true;
    for (string_t::size_type i = 0; i < arg.size(); ++i) {
        if (arg[i] < ' ' || arg[i] >= 127) {
            isPrintable = false;
            break;
        }
    }

    // Format it
    if (arg.empty()) {
        return "\"\"";
    } else if (isPrintable && arg.find_first_of(" \"\'<|${}#") == arg.npos) {
        // Nothing special
        return arg;
    } else if (isPrintable && arg.find_first_of("\"\\$") == arg.npos) {
        // Simple double-quote
        return "\"" + arg + "\"";
    } else if (isPrintable && arg.find_first_of("\'") == arg.npos) {
        // Simple single-quote
        return "'" + arg + "'";
    } else {
        // Full version
        string_t result = "\"";
        for (string_t::size_type i = 0; i < arg.size(); ++i) {
            uint8_t ch = arg[i];
            if (ch == '"' || ch == '\\' || ch == '$') {
                result += '\\';
                result += char(ch);
            } else if (ch == '\r') {
                result += "\\r";
            } else if (ch == '\n') {
                result += "\\n";
            } else if (ch == '\t') {
                result += "\\t";
            } else if (ch == '\0') {
                result += "\\0";
            } else if (ch >= ' ' && ch < 127) {
                result += char(ch);
            } else {
                result += format("\\x%02X", ch);
            }
        }
        result += "\"";
        return result;
    }
}

static string_t
pad(string_t s, string_t::size_type len)
{
    if (s.size() < len) {
        s.append(len - s.size(), ' ');
    }
    return s;
}

namespace {
    class IndirectSorter {
     public:
        IndirectSorter(StringList_t& values)
            : values(values)
            { }
        bool operator()(int32_t a, int32_t b)
            { return values[a] < values[b]; }
     private:
        StringList_t& values;
    };
}

/** Get keys matching a wildcard (redis KEYS command).
    \param db Database to work on
    \param match Wildcard
    \param keys [out] List of keys */
static void
getKeys(DbClient& db, string_t match, StringList_t& keys)
{
    std::auto_ptr<IntValue> result(db.call(DbRequest().withString("KEYS").withString(match)));
    toStringListResult(result.get(), keys);
    std::sort(keys.begin(), keys.end());
}

/** Export a database subtree. */
static void
exportSubtree(std::ostream& out, DbClient& db, string_t match)
{
    StringList_t keys;
    getKeys(db, match, keys);
    for (uint32_t i = 0; i < keys.size(); ++i) {
        const string_t& name = keys[i];
        DbKey k(db, name);
        switch (k.getType()) {
         case DbKey::None:
            out << "# warning: key " << quoteConsoleString(name) << " got deleted during export\n";
            break;
         case DbKey::String:
            out << "silent redis set   "
                << pad(quoteConsoleString(name), 30)
                << " "
                << quoteConsoleString(DbString(db, name).get())
                << "\n";
            break;
         case DbKey::List:
            {
                StringList_t values;
                DbStringList(db, name).getAll(values);
                for (uint32_t j = 0; j < values.size(); ++j) {
                    out << "silent redis rpush "
                        << pad(quoteConsoleString(name), 30)
                        << " "
                        << quoteConsoleString(values[j])
                        << "\n";
                }
            }
            break;
         case DbKey::Set:
            {
                StringList_t values;
                DbStringSet(db, name).getAll(values);
                std::sort(values.begin(), values.end());
                for (uint32_t j = 0; j < values.size(); ++j) {
                    out << "silent redis sadd  "
                        << pad(quoteConsoleString(name), 30)
                        << " "
                        << quoteConsoleString(values[j])
                        << "\n";
                }
            }
            break;
         case DbKey::Hash:
            {
                StringList_t values;
                DbHash(db, name).getAll(values);

                // Sort for reproducability!
                IntList_t indexes;
                for (uint32_t j = 0; j < values.size(); j += 2) {
                    indexes.push_back(j);
                }
                std::sort(indexes.begin(), indexes.end(), IndirectSorter(values));

                // Output
                for (uint32_t j = 0; j < indexes.size(); ++j) {
                    out << "silent redis hset  "
                        << pad(quoteConsoleString(name), 30)
                        << " "
                        << quoteConsoleString(values[indexes[j]])
                        << " "
                        << quoteConsoleString(values[indexes[j]+1])
                        << "\n";
                }
            }
            break;
         case DbKey::ZSet:
         case DbKey::Unknown:
         default:
            out << "# warning: key " << quoteConsoleString(keys[i]) << " has an unsupported type\n";
            break;
        }
    }
}

void
doDatabaseExport(std::ostream& out, IntArgBlock& args)
{
    bool withDelete = false;
    while (args.getNumArgs() > 0) {
        string_t what = toStringResult(args.getNext());
        if (what == "--delete") {
            withDelete = true;
        } else {
            if (withDelete) {
                out << "redis keys " << quoteConsoleString(what) << " | silent noerror redis del\n";
            }
            exportSubtree(out, database_connection, what);
        }
    }
}
