/**
  *  \file planetscentral/rater/difficulty.cc
  *  \brief Rating Game Difficulty
  */

#include <cmath>
#include "planetscentral/rater/difficulty.h"
#include "cpluslib/misc.h"
#include "cpluslib/string.h"
#include "util/conf.h"
#include "io/textfile.h"
#include "game/struct.h"
#include "io/io.h"
#include "io/dirs.h"
#include "util/except.h"

/** Average cost of a ship. A ship list has a difficulty rating of 100% if a ship
    costs on average this many T+D+M+$. */
const double AVG_SHIP_COST = 2000.0;

/** Average minerals on planet. A universe has a difficulty rating of 100% if
    a planet has on average this many T+D+M in core and on surface combined. */
const double AVG_MINERALS = 1800.0;

/** Average natives on planet. A universe has a difficulty rating of 100% if a
    planet has on average this many native clans. */
const double AVG_NATIVES = 25000.0;

/**************************** Shiplist Rating ****************************/

/** Find best engine: first one that does warp 9 at 120% fuel usage. */
static bool
findBestEngine(Directory& dir, TEngine& engine)
{
    try {
        Ptr<Stream> s = dir.openFile("engspec.dat", Stream::C_READ);
        for (int i = 1; i <= NUM_ENGINES; ++i) {
            getStructureT(*s, engine);
            if (engine.fuel[8] <= 120*81) {
                break;
            }
        }
        return true;
    }
    catch (FileProblemException& e) {
        return false;
    }
}

/** Find best beam: best bang. */
static bool
findBestBeam(Directory& dir, TBeam& beam)
{
    try {
        Ptr<Stream> s = dir.openFile("beamspec.dat", Stream::C_READ);
        getStructureT(*s, beam);
        for (int i = 2; i <= NUM_BEAMS; ++i) {
            TBeam b2;
            getStructureT(*s, b2);
            if (b2.kill + b2.damage > beam.kill + beam.damage) {
                beam = b2;
            }
        }
        return true;
    }
    catch (FileProblemException& e) {
        return false;
    }
}

/** Find best torpedo: best bang. */
static bool
findBestTorpedo(Directory& dir, TTorpedo& torp)
{
    try {
        Ptr<Stream> s = dir.openFile("torpspec.dat", Stream::C_READ);
        getStructureT(*s, torp);
        for (int i = 2; i <= NUM_TORPS; ++i) {
            TTorpedo t2;
            getStructureT(*s, t2);
            if (t2.kill + t2.damage > torp.kill + torp.damage) {
                torp = t2;
            }
        }
        return true;
    }
    catch (FileProblemException& e) {
        return false;
    }
}

static int32_t
sumCost(const TCost& c)
{
    return int32_t(c.t) + c.d + c.m + c.mc;
}

/**************************** DifficultyRater ****************************/

/** Constructor. */
DifficultyRater::DifficultyRater()
    : config_known(),
      shiplist_known(false)
{ }

DifficultyRater::~DifficultyRater()
{ }

/** Process a ship list. */
void
DifficultyRater::addShipList(Directory& dir)
{
    // Find best possible equipment
    TEngine bestEngine;
    if (!findBestEngine(dir, bestEngine))
        return;

    TBeam bestBeam;
    if (!findBestBeam(dir, bestBeam))
        return;

    TTorpedo bestTorpedo;
    if (!findBestTorpedo(dir, bestTorpedo))
        return;

    // Process hulls
    try {
        Ptr<Stream> s = dir.openFile("hullspec.dat", Stream::C_READ);
        int32_t totalCost = 0;
        int32_t count = 0;
        THull hull;
        while (getStructure(*s, hull)) {
            totalCost += hull.t + hull.d + hull.m + hull.mc;
            totalCost += hull.engines * sumCost(bestEngine.cost);
            totalCost += hull.beams * sumCost(bestBeam.cost);
            totalCost += hull.launchers * sumCost(bestTorpedo.laun_cost);
            ++count;
        }
        if (count > 0) {
            shiplist_known = true;
            shiplist_average_cost = totalCost / count;
        }
    }
    catch (FileProblemException& e) { }
}

/** Process a set of config files. */
void
DifficultyRater::addConfigDirectory(Directory& dir)
{
    // Process config files
    static const char*const files[] = { "pconfig.src", "shiplist.txt", "amaster.src", "pmaster.cfg" };
    for (uint_t i = 0; i < countof(files); ++i) {
        Ptr<Stream> s = dir.openFileNT(files[i], Stream::C_READ);
        if (s) {
            addConfigFile(*s);
        }
    }
}

/** Process a config file or fragment. */
void
DifficultyRater::addConfigFile(Stream& s)
{
    TextFile tf(s);
    string_t line;
    string_t prefix;
    while (tf.getLine(line)) {
        line = strTrim(line);
        removeComment(line);
        if (line.empty()) {
            // ignore
        } else if (line[0] == '%') {
            // section delimiter
            prefix = strTrim(line.substr(1)) + ".";
        } else {
            string_t::size_type n = line.find('=');
            if (n != line.npos) {
                addConfig(prefix + strRTrim(line.substr(0, n)), strLTrim(line.substr(n+1)));
            }
        }
    }
}

/** Process a config option.
    \param name Option in format "section.key"
    \param value Value */
void
DifficultyRater::addConfig(string_t name, string_t value)
{
    // Mapping of names to config keys
    struct Map {
        const char* name;
        Config what;
    };
    static Map mapping[] = {
        { "amaster.nativerange",               Master_NativeRanges },
        { "amaster.nativesonplanetfrequency",  Master_NativeFrequency },
        { "amaster.planetcorerangesalternate", Master_CoreRangesAlternate },
        { "amaster.planetcorerangesusual",     Master_CoreRangesUsual },
        { "amaster.planetcoreusualfrequency",  Master_CoreUsualFrequency },
        { "amaster.planetsurfaceranges",       Master_SurfaceRanges },
        { "phost.colonisttaxrate",             Host_ColonistTaxRate },
        { "phost.hisseffectrate",              Host_HissEffectRate },            // optional
        { "phost.nativetaxrate",               Host_NativeTaxRate },             // optional
        { "phost.playerrace",                  Host_PlayerRace },                // optional
        { "phost.playerspecialmission",        Host_PlayerSpecialMission },      // optional
        { "phost.productionrate",              Host_ProductionRate },
        { "phost.raceminingrate",              Host_MiningRate },
        { "phost.racetaxrate",                 Host_ColonistTaxRate },           // old name
        { "pmaster.nativeclansrange",          Master_NativeClansRange },
        { "pmaster.nativesonplanetfrequency",  Master_NativeFrequency },
        { "pmaster.planetcorerangesalternate", Master_CoreRangesAlternate },
        { "pmaster.planetcorerangesusual",     Master_CoreRangesUsual },
        { "pmaster.planetcoreusualfrequency",  Master_CoreUsualFrequency },
        { "pmaster.planetsurfaceranges",       Master_SurfaceRanges },
    };

    // Look it up
    uint_t index = Config_MAX;
    for (uint_t i = 0; i < countof(mapping); ++i) {
        if (strCaseCmp(mapping[i].name, name) == 0) {
            index = mapping[i].what;
            break;
        }
    }

    // Found? Assign it.
    if (index != Config_MAX) {
        ValueIntParser().parseArray(value, config_values[index], 11);
        config_known += Config(index);
    }
}

/** Check whether a partial rating is known. */
bool
DifficultyRater::isRatingKnown(Rating whichone) const
{
    switch (whichone) {
     case ShiplistRating:
        return shiplist_known;

     case MineralRating:
        return config_known.contains(Master_CoreRangesUsual)
            && config_known.contains(Master_CoreRangesAlternate)
            && config_known.contains(Master_CoreUsualFrequency)
            && config_known.contains(Master_SurfaceRanges);

     case NativeRating:
        return config_known.contains(Master_NativeFrequency)
            && (config_known.contains(Master_NativeRanges)
                || config_known.contains(Master_NativeClansRange));

     case ProductionRating:
        /* optional: Host_HissEffectRate, Host_NativeTaxRate, Host_PlayerRace, Host_PlayerSpecialMission */
        return config_known.contains(Host_ProductionRate)
            && config_known.contains(Host_MiningRate)
            && config_known.contains(Host_ColonistTaxRate);
    }

    return false;
}

/** Get rating.
    \param whichone partial rating to query */
double
DifficultyRater::getRating(Rating whichone) const
{
    if (!isRatingKnown(whichone)) {
        return 1.00;
    } else {
        switch (whichone) {
         case ShiplistRating:
            return std::pow(double(shiplist_average_cost) / AVG_SHIP_COST, 0.33);

         case MineralRating:
            return std::pow(AVG_MINERALS / getAverageMinerals(), 0.33);

         case NativeRating:
            return std::pow(AVG_NATIVES / getAverageNatives(), 0.33);

         case ProductionRating:
            return 100.0 / getAverageVPI();
        }
        return 1.00;
    }
}

/** Get total game rating. */
double
DifficultyRater::getTotalRating() const
{
    return getRating(ShiplistRating)
        * getRating(MineralRating)
        * getRating(NativeRating)
        * getRating(ProductionRating);
}


/** This computes the average amount of minerals on planets. A game is easier
    when it has many minerals.

    AMASTER distinguishes between usual and alternate minerals. We assume that
    less frequent of the two is the bigger value. AMASTER's selection frequency
    is shifted 3/4 towards the 0%/100% point, to avoid that the exceptional high
    value dominates the regular value (reading: the exceptional high value needs
    extra logistics to use).

    In addition, we add average surface minerals. */
double
DifficultyRater::getAverageMinerals() const
{
    double averageMinerals = 0;

    // Iterate through slots 1..3, i.e. T/D/M; skip slot 0, i.e. N.
    for (int slot = 1; slot <= 3; ++slot) {
        // Core:
        int usual_max = config_values[Master_CoreRangesUsual][slot+4];
        int usual_min = config_values[Master_CoreRangesUsual][slot];
        int alt_max   = config_values[Master_CoreRangesAlternate][slot+4];
        int alt_min   = config_values[Master_CoreRangesAlternate][slot];

        double usual_freq = config_values[Master_CoreUsualFrequency][slot] / 100.0;
        if (usual_freq > 0.5) {
            usual_freq = 1 - (1 - usual_freq)*0.25;
        } else {
            usual_freq *= 0.25;
        }

        averageMinerals += ((usual_max + usual_min) * usual_freq + (alt_max + alt_min) * (1.0 - usual_freq)) / 2.0;

        // Surface:
        int sfc_min = config_values[Master_SurfaceRanges][slot];
        int sfc_max = config_values[Master_SurfaceRanges][slot+4];

        averageMinerals += (sfc_max + sfc_min)/2.0;
    }

    return averageMinerals;
}

/** Computes the average number of natives on a planet, and that's about it.
    The idea is to get an average income, which happens to scale linearly
    with the number of natives.

    This assumes that all natives are equally valuable, i.e. the added
    benefit of Ins/Rep/Bov cancels out Amo, and governments are equally
    distributed. If desired, we could check those, too. */
double
DifficultyRater::getAverageNatives() const
{
    int averageClansX2 = config_known.contains(Master_NativeRanges)
        ? 10*(config_values[Master_NativeRanges][0] + config_values[Master_NativeRanges][1])
        : (config_values[Master_NativeClansRange][0] + config_values[Master_NativeClansRange][1]);
    return config_values[Master_NativeFrequency][0] * averageClansX2 / 200.0;
}

/** Compute average VPI. */
double
DifficultyRater::getAverageVPI() const
{
    int totalProd = 0;
    int totalMini = 0;
    int totalCtax = 0;
    int totalNtax = 0;
    int totalHiss = 0;

    for (int i = 0; i < 11; ++i) {
        // Get config, process defaults
        int prod = config_values[Host_ProductionRate][i];
        int mini = config_values[Host_MiningRate][i];
        int ctax = config_values[Host_ColonistTaxRate][i];
        int ntax = config_known.contains(Host_NativeTaxRate)        ? config_values[Host_NativeTaxRate][i]        : ctax;
        int race = config_known.contains(Host_PlayerRace)           ? config_values[Host_PlayerRace][i]           : i+1;
        int misn = config_known.contains(Host_PlayerSpecialMission) ? config_values[Host_PlayerSpecialMission][i] : race;
        int hiss = misn == 2 ? config_known.contains(Host_HissEffectRate) ? config_values[Host_HissEffectRate][i] : 5 : 0;

        totalProd += prod;
        totalMini += mini;
        totalCtax += ctax;
        totalNtax += ntax;
        totalHiss += hiss;
    }

    // VPI formula
    return -8.0
        + 50*std::pow(1 + (totalProd/11.0 - 100)/100, 0.66)
        + 50*std::sqrt(totalMini/11.0/100)
        + std::sqrt((totalNtax/11.0 + totalCtax/11.0)/2*(1+totalHiss/11.0/5))*0.8;
}
