/**
  *  \file planetscentral/talk/group.cc
  */

#include "planetscentral/talk/group.h"
#include "planetscentral/talk/db.h"
#include "cpluslib/misc.h"
#include "int/array.h"
#include "planetscentral/talk/render.h"

// group:$GRID:header : hash
//         - name                          Name
//         - description                   Description
//         - parent                        If present, GRID of parent group
//         - key                           Sort key for displaying groups
                             
//         Well-known GRIDs:    
//         - root               
//         - active                        Active/joining games
//         - finished                      Finished games

// group:$GRID:groups : set
// group:$GRID:forums : set
//         - GRIDs of subgroups
//         - FIDs of forums

Group::Group(DbClient& client, string_t groupId)
    : group(DbSubtree(client, GROUP_ROOT).subtree(groupId)),
      id(groupId)
{ }

DbHash
Group::header()
{
    return group.hashKey("header");
}

DbHashStringField
Group::name()
{
    return header().stringField("name");
}

DbHashStringField
Group::description()
{
    return header().stringField("description");
}

DbHashStringField
Group::key()
{
    return header().stringField("key");
}

DbHashIntField
Group::unlisted()
{
    return header().intField("unlisted");
}

bool
Group::exists()
{
    return header().exists();
}

DbIntSet
Group::forums()
{
    return group.intSetKey("forums");
}

DbStringSet
Group::subgroups()
{
    return group.stringSetKey("groups");
}

string_t
Group::getParent()
{
    return header().getStringField("parent");
}

void
Group::setParent(string_t newParent)
{
    string_t oldParent = getParent();
    Group oldGroup(group.getClient(), oldParent);
    Group newGroup(group.getClient(), newParent);
    if (oldParent != newParent) {
        if (oldParent.empty()) {
            newGroup.subgroups().add(id);
        } else if (newParent.empty()) {
            oldGroup.subgroups().remove(id);
        } else {
            oldGroup.subgroups().moveTo(id, newGroup.subgroups());
        }
        header().setField("parent", newParent);
    }
}

IntValue*
Group::describe(RenderState& state)
{
    /* @type TalkGroupInfo
       Information about a forum group.

       @key name:Str         (Name)
       @key description:Str  (Description, rendered using {RENDEROPTION})
       @key parent:GRID      (Parent group)
       @key unlisted:Int     (If nonzero, group is unlistable) */

    // FIXME: can we use HMGET?
    const char*const fields[] = { "name", "description", "parent", "unlisted" };
    Ptr<IntArrayData> ad = new IntArrayData();
    for (std::size_t i = 0; i < countof(fields); ++i) {
        ad->content.pushNew(makeStringValue(fields[i]));
        if (fields[i] == "description") {
            ad->content.pushNew(makeStringValue(state.render(description().get(), group.getClient())));
        } else {
            ad->content.pushNew(header().getFieldRaw(fields[i]));
        }
    }
    ad->addDimension(ad->content.size());
    return new IntArray(ad);
}
