/**
  *  \file planetscentral/talk/user.cc
  *  \brief Talk: User profile access
  *
  *  Users have, for our purposes, three identifying names:
  *  - a user Id. This is used everywhere in the database.
  *    Although it is a number, all programs treat it as string.
  *    User Ids taken from the database are trusted. User Ids
  *    are unique and not recycled.
  *  - a login name. This is the name used in URLs. There is an
  *    index mapping login names to user Ids. This is also the
  *    name users use to refer to other users (in at-links, for
  *    example). Login names are unique, but can be recycled.
  *  - a screen name. This one is only displayed and never used
  *    in any machine interface; there is no index and there is
  *    no mechanism to make them unique.
  */

#include "planetscentral/talk/user.h"
#include "planetscentral/talk/db.h"

static string_t
simplifyLoginName(string_t s)
{
    string_t result;
    bool needsp = false;
    for (string_t::size_type i = 0; i < s.size(); ++i) {
        char ch = s[i];
        if ((ch >= 'A' && ch <= 'Z') || (ch >= 'a' && ch <= 'z') || (ch >= '0' && ch <= '9')) {
            if (needsp) {
                result += '_';
            }
            result += std::tolower(ch);
            needsp = false;
        } else {
            needsp = (result.size() != 0);
        }
    }
    return result;
}

/********************************** User *********************************/

/** Constructor.
    \param client Database Client
    \param userId User Id ("1001") */
User::User(DbClient& client, string_t userId)
    : user(DbSubtree(client, USER_ROOT).subtree(userId))
{ }

/** Get user's screen name.
    This value is stored in the user's profile.
    Since 20110901, there is no longer an automatic fallback to the login name,
    all users' screen names are filled in.
    This is also the reason there is no need to implement a fallback to the
    default profile. */
string_t
User::getScreenName()
{
    return profile().getStringField("screenname");
}

/** Get user's login name. This is the name he uses to login, and which others
    use to refer to him. */
string_t
User::getLoginName()
{
    return user.stringKey("name").get();
}

/** Get user's real name. If this user's real name is not available to others,
    returns a null string. */
string_t
User::getRealName()
{
    std::auto_ptr<IntValue> allow(getProfileRaw("inforealnameflag"));
    if (allow.get() != 0 && toIntResult(allow.get()) > 0) {
        return profile().getStringField("realname");
    } else {
        return string_t();
    }
}

/** Get raw value from user profile. */
IntValue*
User::getProfileRaw(string_t key)
{
    IntValue* result = profile().getFieldRaw(key);
    if (!result) {
        result = DbHash(user.getClient(), "default:profile").getFieldRaw(key);
    }
    return result;
}

/** Get PM mail type. */
string_t
User::getPMMailType()
{
    std::auto_ptr<IntValue> value(getProfileRaw("mailpmtype"));
    return toStringResult(value.get());
}

/** Get auto-watch flag. */
bool
User::isAutoWatch()
{
    // Unset means yes
    std::auto_ptr<IntValue> value(getProfileRaw("talkautowatch"));
    return value.get() == 0
        || toIntResult(value.get()) != 0;
}

/** Get watch-individual flag. */
bool
User::isWatchIndividual()
{
    // Unset means no
    std::auto_ptr<IntValue> value(getProfileRaw("talkwatchindividual"));
    return toIntResult(value.get()) > 0;
}

/** Get user's profile. */
DbHash
User::profile()
{
    return user.hashKey("profile");
}

/** Get user's forum data. All data manipulated by the forum is stored here. */
DbSubtree
User::forumData()
{
    return user.subtree("forum");
}

/** Get set of user's posted messages. */
DbIntSet
User::postedMessages()
{
    return forumData().intSetKey("posted");
}

/** Get user's newsrc. \see Newsrc. */
DbSubtree
User::newsrc()
{
    return forumData().subtree("newsrc");
}

/** Get user's personal-message data. */
DbSubtree
User::pmFolderData()
{
    return user.subtree("pm:folder");
}

/** Get user's personal folder counter. */
DbInt
User::pmFolderCount()
{
    return pmFolderData().intKey("id");
}

/** Get list of user's personal folders. */
DbIntSet
User::pmFolders()
{
    return pmFolderData().intSetKey("all");
}

/** Get list of watched forums. */
DbIntSet
User::watchedForums()
{
    return forumData().intSetKey("watchedForums");
}

/** Get list of watched topics. */
DbIntSet
User::watchedTopics()
{
    return forumData().intSetKey("watchedThreads");
}

/** Get list of notified forums (=watched, but notification has been sent). */
DbIntSet
User::notifiedForums()
{
    return forumData().intSetKey("notifiedForums");
}

/** Get list of notified topics (=watched, but notification has been sent). */
DbIntSet
User::notifiedTopics()
{
    return forumData().intSetKey("notifiedThreads");
}

/** Map login name to user Id.
    \param db Database Client
    \param login Login name
    \return user Id; empty string if invalid */
string_t
User::getUserIdFromLogin(DbClient& db, string_t login)
{
    string_t simplifiedLogin = simplifyLoginName(login);
    if (simplifiedLogin.empty()) {
        // Name consists of illegal characters only
        return string_t();
    }
    string_t userId = DbSubtree(db, "uid:").stringKey(simplifiedLogin).get();
    if (userId.find_first_not_of("0") == string_t::npos) {
        // does not exist, return empty string
        return string_t();
    }
    return userId;
}
